"""Tests for tool registry."""

import asyncio

import pytest

from agenium.tools.registry import ToolContext, ToolRegistry


@pytest.fixture
def registry():
    return ToolRegistry()


class TestToolRegistry:
    def test_register(self, registry: ToolRegistry):
        async def handler(name: str) -> str:
            return f"hi {name}"

        defn = registry.register("greet", handler, description="Greet someone")
        assert defn.name == "greet"
        assert defn.description == "Greet someone"
        assert len(registry) == 1

    def test_unregister(self, registry: ToolRegistry):
        async def handler() -> str:
            return "ok"

        registry.register("test", handler)
        assert registry.unregister("test") is True
        assert registry.unregister("test") is False
        assert len(registry) == 0

    def test_list_tools(self, registry: ToolRegistry):
        async def h1() -> str:
            return "a"

        async def h2() -> str:
            return "b"

        registry.register("a", h1)
        registry.register("b", h2)
        tools = registry.list_tools()
        assert len(tools) == 2
        names = {t.name for t in tools}
        assert names == {"a", "b"}

    def test_contains(self, registry: ToolRegistry):
        async def handler() -> str:
            return "ok"

        registry.register("exists", handler)
        assert "exists" in registry
        assert "missing" not in registry

    @pytest.mark.asyncio
    async def test_invoke_success(self, registry: ToolRegistry):
        async def add(a: int, b: int) -> int:
            return a + b

        registry.register("add", add)
        ctx = ToolContext(session_id="s1", agent_name="test")
        result = await registry.invoke("add", {"a": 2, "b": 3}, ctx)
        assert result.success is True
        assert result.result == 5

    @pytest.mark.asyncio
    async def test_invoke_not_found(self, registry: ToolRegistry):
        ctx = ToolContext(session_id="s1", agent_name="test")
        result = await registry.invoke("missing", {}, ctx)
        assert result.success is False
        assert "not found" in result.error.lower()

    @pytest.mark.asyncio
    async def test_invoke_error(self, registry: ToolRegistry):
        async def fail() -> None:
            raise ValueError("boom")

        registry.register("fail", fail)
        ctx = ToolContext(session_id="s1", agent_name="test")
        result = await registry.invoke("fail", {}, ctx)
        assert result.success is False
        assert "boom" in result.error



"""
AGENIUM Core Types

Fundamental data structures for agent-to-agent communication.
"""

from __future__ import annotations

import re
import time
import uuid
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional


# ============================================================================
# Agent Identity
# ============================================================================


@dataclass
class AgentID:
    """Unique identifier for an agent in the network."""

    name: str
    """Unique name (used in agent:// URI)."""

    public_key: str
    """Ed25519 public key (base64 encoded)."""

    description: str = ""
    """Optional human-readable description."""


@dataclass
class AgentToolRef:
    """A tool/function that an agent exposes."""

    name: str
    description: str = ""


@dataclass
class AgentEndpoint:
    """Resolved endpoint for connecting to an agent."""

    agent_id: AgentID
    url: str
    cert_fingerprint: Optional[str] = None
    protocol_versions: list[str] = field(default_factory=lambda: ["1.0"])
    capabilities: list[str] = field(default_factory=list)
    tools: list[AgentToolRef] = field(default_factory=list)
    ttl: int = 300
    resolved_at: float = field(default_factory=time.time)


# ============================================================================
# Session State Machine
# ============================================================================


class SessionState(str, Enum):
    """Session lifecycle states."""

    INITIATING = "initiating"
    HANDSHAKE = "handshake"
    ACTIVE = "active"
    SUSPENDED = "suspended"
    RESUMING = "resuming"
    CLOSING = "closing"
    CLOSED = "closed"
    ERROR = "error"


@dataclass
class Session:
    """An active session between two agents."""

    id: str
    local_agent: AgentID
    remote_agent: AgentID
    state: SessionState = SessionState.INITIATING
    created_at: float = field(default_factory=time.time)
    updated_at: float = field(default_factory=time.time)
    metadata: dict[str, Any] = field(default_factory=dict)


# ============================================================================
# URI Parsing
# ============================================================================

_AGENT_NAME_RE = re.compile(r"^[a-z0-9\u0080-\uffff]([a-z0-9\u0080-\uffff-]*[a-z0-9\u0080-\uffff])?$")


def parse_agent_uri(uri: str) -> Optional[str]:
    """
    Parse agent:// URI and return the agent name.

    Format: agent://name
    - Length: 2-50 characters
    - Characters: lowercase alphanumeric, hyphens, unicode (IDN)
    - Must not start or end with hyphen

    Returns None if invalid.

    >>> parse_agent_uri("agent://alice")
    'alice'
    >>> parse_agent_uri("agent://my-agent")
    'my-agent'
    >>> parse_agent_uri("invalid") is None
    True
    """
    if not uri.startswith("agent://"):
        return None

    name = uri[8:].lower()

    if len(name) < 2 or len(name) > 50:
        return None

    if name.startswith("-") or name.endswith("-"):
        return None

    if not _AGENT_NAME_RE.match(name):
        return None

    return name


def is_valid_agent_uri(uri: str) -> bool:
    """Check if a string is a valid agent:// URI."""
    return parse_agent_uri(uri) is not None


def validate_agent_name(name: str) -> bool:
    """
    Validate an agent name (without the agent:// prefix).

    >>> validate_agent_name("alice")
    True
    >>> validate_agent_name("a")
    False
    """
    if len(name) < 2 or len(name) > 50:
        return False
    if name.startswith("-") or name.endswith("-"):
        return False
    return bool(_AGENT_NAME_RE.match(name.lower()))


def to_agent_uri(name: str) -> str:
    """Convert a name to agent:// URI format."""
    return f"agent://{name.lower()}"


def generate_id() -> str:
    """Generate a unique ID."""
    return str(uuid.uuid4())



"""
AGENIUM Protocol Types

Frame types for agent-to-agent messaging.
"""

from __future__ import annotations

import time
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional

from ..core.types import generate_id


class MessageType(str, Enum):
    REQUEST = "request"
    RESPONSE = "response"
    EVENT = "event"
    ERROR = "error"


class ErrorCodes:
    INVALID_REQUEST = -32600
    METHOD_NOT_FOUND = -32601
    INVALID_PARAMS = -32602
    INTERNAL_ERROR = -32603
    TIMEOUT = -32000
    SESSION_NOT_FOUND = -32001
    DUPLICATE_MESSAGE = -32002


@dataclass
class RequestFrame:
    """A request expecting a response."""

    type: MessageType = MessageType.REQUEST
    id: str = field(default_factory=generate_id)
    method: str = ""
    params: dict[str, Any] = field(default_factory=dict)
    session_id: Optional[str] = None
    timestamp: float = field(default_factory=time.time)


@dataclass
class ResponseFrame:
    """Response to a request."""

    type: MessageType = MessageType.RESPONSE
    id: str = field(default_factory=generate_id)
    request_id: str = ""
    result: Any = None
    session_id: Optional[str] = None
    timestamp: float = field(default_factory=time.time)


@dataclass
class EventFrame:
    """One-way event (no response expected)."""

    type: MessageType = MessageType.EVENT
    id: str = field(default_factory=generate_id)
    event: str = ""
    data: Any = None
    session_id: Optional[str] = None
    timestamp: float = field(default_factory=time.time)


@dataclass
class ErrorFrame:
    """Error response."""

    type: MessageType = MessageType.ERROR
    id: str = field(default_factory=generate_id)
    request_id: str = ""
    code: int = ErrorCodes.INTERNAL_ERROR
    message: str = ""
    data: Any = None
    session_id: Optional[str] = None
    timestamp: float = field(default_factory=time.time)


AnyFrame = RequestFrame | ResponseFrame | EventFrame | ErrorFrame


def create_request_frame(
    method: str,
    params: dict[str, Any] | None = None,
    session_id: str | None = None,
) -> RequestFrame:
    """Create a new request frame."""
    return RequestFrame(method=method, params=params or {}, session_id=session_id)


def create_response_frame(
    request_id: str,
    result: Any = None,
    session_id: str | None = None,
) -> ResponseFrame:
    """Create a response frame for a request."""
    return ResponseFrame(request_id=request_id, result=result, session_id=session_id)


def create_event_frame(
    event: str,
    data: Any = None,
    session_id: str | None = None,
) -> EventFrame:
    """Create a new event frame."""
    return EventFrame(event=event, data=data, session_id=session_id)


def create_error_frame(
    request_id: str,
    code: int = ErrorCodes.INTERNAL_ERROR,
    message: str = "",
    data: Any = None,
    session_id: str | None = None,
) -> ErrorFrame:
    """Create an error response frame."""
    return ErrorFrame(
        request_id=request_id, code=code, message=message, data=data, session_id=session_id
    )


def validate_frame(frame: AnyFrame) -> bool:
    """Validate a frame has required fields."""
    if not isinstance(frame, (RequestFrame, ResponseFrame, EventFrame, ErrorFrame)):
        return False
    if not frame.id:
        return False
    if isinstance(frame, RequestFrame) and not frame.method:
        return False
    if isinstance(frame, ResponseFrame) and not frame.request_id:
        return False
    if isinstance(frame, EventFrame) and not frame.event:
        return False
    if isinstance(frame, ErrorFrame) and not frame.request_id:
        return False
    return True



"""
Tool Registry

Register, list, and invoke tools on an agent.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, Optional


ToolHandler = Callable[..., Awaitable[Any]]


@dataclass
class ToolDefinition:
    """Definition of a tool that an agent can expose."""

    name: str
    description: str = ""
    input_schema: Optional[dict[str, Any]] = None
    output_schema: Optional[dict[str, Any]] = None


@dataclass
class ToolContext:
    """Context passed to tool handlers."""

    session_id: str
    agent_name: str
    metadata: dict[str, Any] = field(default_factory=dict)


@dataclass
class ToolInvokeResult:
    """Result of invoking a tool."""

    success: bool
    result: Any = None
    error: Optional[str] = None


class ToolRegistry:
    """Registry for agent tools."""

    def __init__(self) -> None:
        self._tools: dict[str, tuple[ToolDefinition, ToolHandler]] = {}

    def register(
        self,
        name: str,
        handler: ToolHandler,
        description: str = "",
        input_schema: dict[str, Any] | None = None,
        output_schema: dict[str, Any] | None = None,
    ) -> ToolDefinition:
        """Register a tool."""
        defn = ToolDefinition(
            name=name,
            description=description,
            input_schema=input_schema,
            output_schema=output_schema,
        )
        self._tools[name] = (defn, handler)
        return defn

    def unregister(self, name: str) -> bool:
        """Unregister a tool. Returns True if it existed."""
        return self._tools.pop(name, None) is not None

    def get(self, name: str) -> tuple[ToolDefinition, ToolHandler] | None:
        """Get a tool definition and handler by name."""
        return self._tools.get(name)

    def list_tools(self) -> list[ToolDefinition]:
        """List all registered tools."""
        return [defn for defn, _ in self._tools.values()]

    async def invoke(
        self, name: str, params: dict[str, Any], context: ToolContext
    ) -> ToolInvokeResult:
        """
        Invoke a tool by name.

        Returns ToolInvokeResult with success/result or error.
        """
        entry = self._tools.get(name)
        if entry is None:
            return ToolInvokeResult(success=False, error=f"Tool not found: {name}")

        _, handler = entry
        try:
            result = await handler(**params)
            return ToolInvokeResult(success=True, result=result)
        except TypeError as e:
            return ToolInvokeResult(success=False, error=f"Invalid parameters: {e}")
        except Exception as e:
            return ToolInvokeResult(success=False, error=f"Tool error: {e}")

    def __len__(self) -> int:
        return len(self._tools)

    def __contains__(self, name: str) -> bool:
        return name in self._tools



"""Tests for crypto module."""

from agenium.crypto.keys import KeyPair, generate_keypair


class TestKeyGeneration:
    def test_generate(self):
        kp = generate_keypair()
        assert isinstance(kp, KeyPair)
        assert kp.public_key_b64
        assert len(kp.public_key_b64) == 44  # 32 bytes base64

    def test_sign_verify(self):
        kp = generate_keypair()
        data = b"hello agenium"
        sig = kp.sign(data)
        assert kp.verify(sig, data) is True

    def test_verify_wrong_data(self):
        kp = generate_keypair()
        sig = kp.sign(b"original")
        assert kp.verify(sig, b"tampered") is False

    def test_pem_export(self):
        kp = generate_keypair()
        assert kp.private_key_pem().startswith("-----BEGIN PRIVATE KEY-----")
        assert kp.public_key_pem().startswith("-----BEGIN PUBLIC KEY-----")

    def test_unique_keys(self):
        k1 = generate_keypair()
        k2 = generate_keypair()
        assert k1.public_key_b64 != k2.public_key_b64



"""
DNS Resolver

Resolves agent:// URIs to endpoints via the AGENIUM DNS system.
"""

from __future__ import annotations

import time
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional

import httpx

from ..core.types import parse_agent_uri, validate_agent_name


# ============================================================================
# Types
# ============================================================================


@dataclass
class AgentTool:
    """A tool advertised by an agent in DNS."""

    name: str
    description: str = ""
    input_schema: Optional[dict[str, Any]] = None
    output_schema: Optional[dict[str, Any]] = None


@dataclass
class ResolvedAgent:
    """Result of resolving an agent name via DNS."""

    name: str
    endpoint: str
    public_key: str
    tools: list[AgentTool] = field(default_factory=list)
    capabilities: list[str] = field(default_factory=list)
    ttl: int = 300
    resolved_at: float = field(default_factory=time.time)
    metadata: dict[str, Any] = field(default_factory=dict)

    @property
    def uri(self) -> str:
        return f"agent://{self.name}"


class DNSErrorCode(str, Enum):
    NOT_FOUND = "NOT_FOUND"
    INVALID_NAME = "INVALID_NAME"
    TIMEOUT = "TIMEOUT"
    SERVER_ERROR = "SERVER_ERROR"
    NETWORK_ERROR = "NETWORK_ERROR"


class DNSError(Exception):
    """Error during DNS resolution."""

    def __init__(self, code: DNSErrorCode, message: str):
        self.code = code
        super().__init__(f"[{code.value}] {message}")


# ============================================================================
# Cache
# ============================================================================


@dataclass
class _CacheEntry:
    agent: ResolvedAgent
    expires_at: float


# ============================================================================
# Resolver
# ============================================================================


@dataclass
class DNSResolverConfig:
    server: str = "185.204.169.26"
    timeout_ms: int = 10_000
    default_ttl_seconds: int = 300
    use_https: bool = False
    port: int = 3000


class DNSResolver:
    """
    Resolve agent:// URIs to endpoints via the AGENIUM DNS system.

    Usage:
        resolver = DNSResolver()
        agent = await resolver.resolve("my-agent")
        print(agent.endpoint)
    """

    def __init__(
        self,
        server: str = "185.204.169.26",
        port: int = 3000,
        timeout_ms: int = 10_000,
        use_https: bool = False,
    ):
        self._config = DNSResolverConfig(
            server=server,
            port=port,
            timeout_ms=timeout_ms,
            use_https=use_https,
        )
        self._cache: dict[str, _CacheEntry] = {}
        self._base_url = (
            f"{'https' if use_https else 'http'}://{server}:{port}"
        )

    async def resolve(self, name: str) -> ResolvedAgent:
        """
        Resolve an agent name to its endpoint.

        Args:
            name: Agent name (e.g. "my-agent") or URI ("agent://my-agent")

        Returns:
            ResolvedAgent with endpoint, tools, etc.

        Raises:
            DNSError on failure
        """
        # Handle both name and URI
        if name.startswith("agent://"):
            parsed = parse_agent_uri(name)
            if parsed is None:
                raise DNSError(DNSErrorCode.INVALID_NAME, f"Invalid URI: {name}")
            name = parsed

        if not validate_agent_name(name):
            raise DNSError(DNSErrorCode.INVALID_NAME, f"Invalid agent name: {name}")

        # Check cache
        cached = self._cache.get(name)
        if cached and cached.expires_at > time.time():
            return cached.agent

        # Resolve via HTTP
        try:
            async with httpx.AsyncClient(timeout=self._config.timeout_ms / 1000) as client:
                resp = await client.get(f"{self._base_url}/dns/lookup/{name}")

                if resp.status_code == 404:
                    raise DNSError(DNSErrorCode.NOT_FOUND, f"Agent not found: {name}")

                if resp.status_code != 200:
                    raise DNSError(
                        DNSErrorCode.SERVER_ERROR,
                        f"DNS server error: {resp.status_code}",
                    )

                data = resp.json()

        except httpx.TimeoutException:
            raise DNSError(DNSErrorCode.TIMEOUT, f"DNS lookup timed out for {name}")
        except httpx.ConnectError as e:
            raise DNSError(DNSErrorCode.NETWORK_ERROR, f"Cannot reach DNS server: {e}")
        except DNSError:
            raise
        except Exception as e:
            raise DNSError(DNSErrorCode.SERVER_ERROR, f"Unexpected error: {e}")

        # Parse response
        agent_data = data.get("agent", data)
        tools_raw = agent_data.get("tools", [])
        tools = [
            AgentTool(
                name=t.get("name", ""),
                description=t.get("description", ""),
                input_schema=t.get("inputSchema") or t.get("input_schema"),
                output_schema=t.get("outputSchema") or t.get("output_schema"),
            )
            for t in tools_raw
        ]

        agent = ResolvedAgent(
            name=name,
            endpoint=agent_data.get("endpoint", ""),
            public_key=agent_data.get("publicKey", agent_data.get("public_key", "")),
            tools=tools,
            capabilities=agent_data.get("capabilities", []),
            ttl=agent_data.get("ttl", self._config.default_ttl_seconds),
            metadata={
                k: v
                for k, v in agent_data.items()
                if k not in ("name", "endpoint", "publicKey", "public_key", "tools", "capabilities", "ttl")
            },
        )

        # Cache
        self._cache[name] = _CacheEntry(
            agent=agent,
            expires_at=time.time() + agent.ttl,
        )

        return agent

    async def resolve_uri(self, uri: str) -> ResolvedAgent:
        """Resolve a full agent:// URI."""
        return await self.resolve(uri)

    def clear_cache(self) -> None:
        """Clear the DNS cache."""
        self._cache.clear()

    def cache_size(self) -> int:
        """Number of entries in cache."""
        return len(self._cache)



"""
AGENIUM — Local, Stateful Agent-to-Agent Communication SDK

Usage:
    from agenium import Agent, DNSResolver, AgentConfig
"""

__version__ = "0.1.0"

from .agent import Agent, AgentConfig, ConnectResult, DNSRegistrationResult
from .dns import DNSResolver, ResolvedAgent, AgentTool
from .protocol import (
    MessageType,
    RequestFrame,
    ResponseFrame,
    EventFrame,
    ErrorFrame,
    create_request_frame,
    create_response_frame,
    create_event_frame,
    create_error_frame,
)
from .tools import ToolRegistry, ToolDefinition
from .core import (
    AgentID,
    AgentEndpoint,
    SessionState,
    parse_agent_uri,
    validate_agent_name,
    to_agent_uri,
)

__all__ = [
    # Agent
    "Agent",
    "AgentConfig",
    "ConnectResult",
    "DNSRegistrationResult",
    # DNS
    "DNSResolver",
    "ResolvedAgent",
    "AgentTool",
    # Protocol
    "MessageType",
    "RequestFrame",
    "ResponseFrame",
    "EventFrame",
    "ErrorFrame",
    "create_request_frame",
    "create_response_frame",
    "create_event_frame",
    "create_error_frame",
    # Tools
    "ToolRegistry",
    "ToolDefinition",
    # Core
    "AgentID",
    "AgentEndpoint",
    "SessionState",
    "parse_agent_uri",
    "validate_agent_name",
    "to_agent_uri",
]



"""Tool system for registering and invoking agent capabilities."""

from .registry import ToolRegistry, ToolDefinition, ToolHandler, ToolContext, ToolInvokeResult

__all__ = ["ToolRegistry", "ToolDefinition", "ToolHandler", "ToolContext", "ToolInvokeResult"]



"""
AGENIUM Agent

High-level agent API with DNS registration, tool system, and messaging.
"""

from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, Optional

import httpx

from .core.types import AgentID, Session, SessionState, generate_id, validate_agent_name
from .crypto.keys import KeyPair, generate_keypair
from .dns.resolver import DNSResolver, ResolvedAgent
from .protocol.types import (
    EventFrame,
    RequestFrame,
    create_event_frame,
    create_request_frame,
)
from .tools.registry import ToolContext, ToolDefinition, ToolHandler, ToolInvokeResult, ToolRegistry

logger = logging.getLogger("agenium")

DEFAULT_DNS_SERVER = "185.204.169.26"
DEFAULT_DNS_PORT = 3000


@dataclass
class AgentConfig:
    """Configuration for an Agent."""

    dns_server: str = DEFAULT_DNS_SERVER
    dns_port: int = DEFAULT_DNS_PORT
    data_dir: str = "~/.agenium"
    connection_timeout_ms: int = 30_000
    request_timeout_ms: int = 30_000
    persistence: bool = True


@dataclass
class ConnectResult:
    """Result of connecting to another agent."""

    success: bool
    session: Optional[Session] = None
    error: Optional[str] = None


@dataclass
class DNSRegistrationResult:
    """Result of DNS registration."""

    success: bool
    domain: Optional[str] = None
    tools: int = 0
    error: Optional[str] = None


class Agent:
    """
    AGENIUM Agent — high-level API for agent-to-agent communication.

    Usage:
        agent = Agent("my-agent")

        @agent.tool("greet", description="Greet someone")
        async def greet(name: str) -> str:
            return f"Hello, {name}!"

        await agent.start(port=8443)
        await agent.register(api_key="dom_xxx")
    """

    def __init__(self, name: str, config: AgentConfig | None = None):
        if not validate_agent_name(name):
            raise ValueError(
                f"Invalid agent name: {name!r}. "
                "Must be 2-50 chars, lowercase alphanumeric/hyphens, no leading/trailing hyphen."
            )

        self._name = name
        self._config = config or AgentConfig()
        self._keys = generate_keypair()
        self._identity = AgentID(
            name=name,
            public_key=self._keys.public_key_b64,
            description=f"AGENIUM Agent: {name}",
        )
        self._tool_registry = ToolRegistry()
        self._resolver = DNSResolver(
            server=self._config.dns_server,
            port=self._config.dns_port,
        )
        self._sessions: dict[str, Session] = {}
        self._dns_api_key: str | None = None
        self._is_running = False
        self._request_handlers: dict[str, Callable] = {}
        self._event_handlers: dict[str, Callable] = {}

    # ========================================================================
    # Properties
    # ========================================================================

    @property
    def name(self) -> str:
        """Agent name."""
        return self._name

    @property
    def uri(self) -> str:
        """Agent URI (agent://name)."""
        return f"agent://{self._name}"

    @property
    def identity(self) -> AgentID:
        """Agent identity."""
        return self._identity

    @property
    def keys(self) -> KeyPair:
        """Agent key pair."""
        return self._keys

    @property
    def is_running(self) -> bool:
        """Whether the agent is running."""
        return self._is_running

    @property
    def sessions(self) -> dict[str, Session]:
        """Active sessions."""
        return dict(self._sessions)

    @property
    def tools(self) -> list[ToolDefinition]:
        """Registered tools."""
        return self._tool_registry.list_tools()

    # ========================================================================
    # Tool Registration
    # ========================================================================

    def tool(
        self,
        name: str,
        description: str = "",
        input_schema: dict[str, Any] | None = None,
        output_schema: dict[str, Any] | None = None,
    ) -> Callable:
        """
        Decorator to register a tool.

        Usage:
            @agent.tool("greet", description="Greet someone")
            async def greet(name: str) -> str:
                return f"Hello, {name}!"
        """

        def decorator(fn: ToolHandler) -> ToolHandler:
            self._tool_registry.register(
                name=name,
                handler=fn,
                description=description,
                input_schema=input_schema,
                output_schema=output_schema,
            )
            return fn

        return decorator

    def register_tool(
        self,
        name: str,
        handler: ToolHandler,
        description: str = "",
        input_schema: dict[str, Any] | None = None,
        output_schema: dict[str, Any] | None = None,
    ) -> ToolDefinition:
        """Register a tool programmatically."""
        return self._tool_registry.register(
            name=name,
            handler=handler,
            description=description,
            input_schema=input_schema,
            output_schema=output_schema,
        )

    async def invoke_tool(
        self, name: str, params: dict[str, Any], session_id: str = ""
    ) -> ToolInvokeResult:
        """Invoke a local tool."""
        ctx = ToolContext(session_id=session_id, agent_name=self._name)
        return await self._tool_registry.invoke(name, params, ctx)

    # ========================================================================
    # Lifecycle
    # ========================================================================

    async def start(self, port: int = 8443, host: str = "0.0.0.0") -> None:
        """
        Start the agent.

        In this initial release, this initializes the agent state.
        Full HTTP/2+mTLS server will be added in a future version.
        """
        if self._is_running:
            return

        self._is_running = True
        logger.info("Agent %s started (port=%d)", self._name, port)

    async def stop(self) -> None:
        """Gracefully stop the agent."""
        if not self._is_running:
            return

        # Close all sessions
        for session in self._sessions.values():
            session.state = SessionState.CLOSED

        self._sessions.clear()
        self._is_running = False
        logger.info("Agent %s stopped", self._name)

    # ========================================================================
    # DNS Registration
    # ========================================================================

    async def register(
        self,
        api_key: str,
        host: str | None = None,
    ) -> DNSRegistrationResult:
        """
        Register this agent on the AGENIUM DNS system.

        Args:
            api_key: DNS API key (format: dom_xxx)
            host: Override endpoint host (auto-detected if not provided)
        """
        self._dns_api_key = api_key

        tools_payload = [
            {
                "name": t.name,
                "description": t.description,
                **({"inputSchema": t.input_schema} if t.input_schema else {}),
            }
            for t in self._tool_registry.list_tools()
        ]

        dns_url = (
            f"http://{self._config.dns_server}:{self._config.dns_port}"
        )

        try:
            async with httpx.AsyncClient(timeout=10) as client:
                resp = await client.post(
                    f"{dns_url}/dns/register",
                    json={
                        "name": self._name,
                        "endpoint": host or "auto",
                        "publicKey": self._keys.public_key_b64,
                        "tools": tools_payload,
                    },
                    headers={"Authorization": f"Bearer {api_key}"},
                )

                if resp.status_code in (200, 201):
                    data = resp.json()
                    return DNSRegistrationResult(
                        success=True,
                        domain=f"agent://{self._name}",
                        tools=len(tools_payload),
                    )
                else:
                    return DNSRegistrationResult(
                        success=False,
                        error=f"Registration failed: {resp.status_code} {resp.text}",
                    )

        except Exception as e:
            return DNSRegistrationResult(success=False, error=str(e))

    # ========================================================================
    # Connection
    # ========================================================================

    async def resolve(self, target: str) -> ResolvedAgent:
        """Resolve an agent name or URI to its endpoint."""
        return await self._resolver.resolve(target)

    async def connect(self, target: str) -> ConnectResult:
        """
        Connect to another agent.

        Args:
            target: Agent name ("my-agent") or URI ("agent://my-agent")

        Returns:
            ConnectResult with session info
        """
        try:
            resolved = await self._resolver.resolve(target)
        except Exception as e:
            return ConnectResult(success=False, error=f"DNS resolution failed: {e}")

        session = Session(
            id=generate_id(),
            local_agent=self._identity,
            remote_agent=AgentID(
                name=resolved.name,
                public_key=resolved.public_key,
            ),
            state=SessionState.ACTIVE,
        )

        self._sessions[session.id] = session
        return ConnectResult(success=True, session=session)

    # ========================================================================
    # Messaging
    # ========================================================================

    async def send(
        self, session_id: str, event: str, data: Any = None
    ) -> bool:
        """Send an event to a session."""
        session = self._sessions.get(session_id)
        if not session:
            logger.warning("Session not found: %s", session_id)
            return False

        frame = create_event_frame(event=event, data=data, session_id=session_id)
        # In this initial release, frame is created but transport is stub
        logger.debug("Event sent: %s → %s", event, session.remote_agent.name)
        return True

    async def call_tool(
        self,
        session_id: str,
        tool_name: str,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """
        Call a tool on a remote agent.

        Args:
            session_id: Active session ID
            tool_name: Name of the remote tool
            params: Tool parameters

        Returns:
            Tool result

        Raises:
            RuntimeError if session not found or call fails
        """
        session = self._sessions.get(session_id)
        if not session:
            raise RuntimeError(f"Session not found: {session_id}")

        frame = create_request_frame(
            method="tool.invoke",
            params={"tool": tool_name, "input": params or {}},
            session_id=session_id,
        )

        # Stub: in full implementation, this sends via mTLS transport
        logger.debug(
            "Tool call: %s.%s(%s)", session.remote_agent.name, tool_name, params
        )
        return None

    # ========================================================================
    # Event Handlers
    # ========================================================================

    def on_request(self, method: str, handler: Callable) -> None:
        """Register a handler for incoming requests."""
        self._request_handlers[method] = handler

    def on_event(self, event: str, handler: Callable) -> None:
        """Register a handler for incoming events."""
        self._event_handlers[event] = handler

    # ========================================================================
    # Repr
    # ========================================================================

    def __repr__(self) -> str:
        state = "running" if self._is_running else "stopped"
        return f"Agent({self._name!r}, {state}, tools={len(self._tool_registry)})"



"""Tests for the Agent class."""

import pytest

from agenium.agent import Agent, AgentConfig


class TestAgentCreation:
    def test_create(self):
        agent = Agent("test-agent")
        assert agent.name == "test-agent"
        assert agent.uri == "agent://test-agent"
        assert not agent.is_running

    def test_invalid_name(self):
        with pytest.raises(ValueError):
            Agent("a")  # too short

    def test_with_config(self):
        config = AgentConfig(dns_server="1.2.3.4", dns_port=5000)
        agent = Agent("my-agent", config=config)
        assert agent.name == "my-agent"

    def test_repr(self):
        agent = Agent("test-agent")
        assert "test-agent" in repr(agent)
        assert "stopped" in repr(agent)


class TestToolDecorator:
    def test_register_tool(self):
        agent = Agent("test-agent")

        @agent.tool("greet", description="Greet someone")
        async def greet(name: str) -> str:
            return f"Hello, {name}!"

        tools = agent.tools
        assert len(tools) == 1
        assert tools[0].name == "greet"
        assert tools[0].description == "Greet someone"

    def test_register_multiple_tools(self):
        agent = Agent("test-agent")

        @agent.tool("a")
        async def tool_a() -> str:
            return "a"

        @agent.tool("b")
        async def tool_b() -> str:
            return "b"

        assert len(agent.tools) == 2


class TestAgentLifecycle:
    @pytest.mark.asyncio
    async def test_start_stop(self):
        agent = Agent("test-agent")
        assert not agent.is_running

        await agent.start()
        assert agent.is_running

        await agent.stop()
        assert not agent.is_running

    @pytest.mark.asyncio
    async def test_double_start(self):
        agent = Agent("test-agent")
        await agent.start()
        await agent.start()  # should not error
        assert agent.is_running
        await agent.stop()

    @pytest.mark.asyncio
    async def test_invoke_local_tool(self):
        agent = Agent("test-agent")

        @agent.tool("add")
        async def add(a: int, b: int) -> int:
            return a + b

        result = await agent.invoke_tool("add", {"a": 3, "b": 4})
        assert result.success
        assert result.result == 7


class TestAgentCrypto:
    def test_has_keys(self):
        agent = Agent("test-agent")
        assert agent.keys.public_key_b64
        assert len(agent.keys.public_key_b64) > 10

    def test_unique_keys(self):
        a1 = Agent("agent-one")
        a2 = Agent("agent-two")
        assert a1.keys.public_key_b64 != a2.keys.public_key_b64



"""Core types and utilities for AGENIUM."""

from .types import (
    AgentID,
    AgentEndpoint,
    AgentToolRef,
    SessionState,
    Session,
    parse_agent_uri,
    validate_agent_name,
    to_agent_uri,
    is_valid_agent_uri,
    generate_id,
)

__all__ = [
    "AgentID",
    "AgentEndpoint",
    "AgentToolRef",
    "SessionState",
    "Session",
    "parse_agent_uri",
    "validate_agent_name",
    "to_agent_uri",
    "is_valid_agent_uri",
    "generate_id",
]



"""Tests for core types and utilities."""

from agenium.core.types import (
    AgentID,
    Session,
    SessionState,
    generate_id,
    is_valid_agent_uri,
    parse_agent_uri,
    to_agent_uri,
    validate_agent_name,
)


class TestParseAgentURI:
    def test_valid_uri(self):
        assert parse_agent_uri("agent://alice") == "alice"

    def test_valid_uri_with_hyphens(self):
        assert parse_agent_uri("agent://my-agent") == "my-agent"

    def test_valid_uri_with_numbers(self):
        assert parse_agent_uri("agent://bot123") == "bot123"

    def test_invalid_no_prefix(self):
        assert parse_agent_uri("alice") is None

    def test_invalid_too_short(self):
        assert parse_agent_uri("agent://a") is None

    def test_invalid_leading_hyphen(self):
        assert parse_agent_uri("agent://-bad") is None

    def test_invalid_trailing_hyphen(self):
        assert parse_agent_uri("agent://bad-") is None

    def test_case_insensitive(self):
        assert parse_agent_uri("agent://MyAgent") == "myagent"


class TestValidateAgentName:
    def test_valid(self):
        assert validate_agent_name("alice") is True
        assert validate_agent_name("my-agent") is True
        assert validate_agent_name("bot123") is True

    def test_invalid(self):
        assert validate_agent_name("a") is False
        assert validate_agent_name("-bad") is False
        assert validate_agent_name("") is False
        assert validate_agent_name("a" * 51) is False


class TestToAgentURI:
    def test_basic(self):
        assert to_agent_uri("alice") == "agent://alice"

    def test_uppercase_lowered(self):
        assert to_agent_uri("Alice") == "agent://alice"


class TestIsValidAgentURI:
    def test_valid(self):
        assert is_valid_agent_uri("agent://alice") is True

    def test_invalid(self):
        assert is_valid_agent_uri("http://alice") is False


class TestGenerateID:
    def test_unique(self):
        ids = {generate_id() for _ in range(100)}
        assert len(ids) == 100

    def test_format(self):
        id_ = generate_id()
        assert len(id_) == 36  # UUID format


class TestSession:
    def test_create(self):
        agent = AgentID(name="test", public_key="abc")
        session = Session(id="s1", local_agent=agent, remote_agent=agent)
        assert session.state == SessionState.INITIATING
        assert session.id == "s1"



"""Message protocol types and frame builders."""

from .types import (
    MessageType,
    RequestFrame,
    ResponseFrame,
    EventFrame,
    ErrorFrame,
    ErrorCodes,
    create_request_frame,
    create_response_frame,
    create_event_frame,
    create_error_frame,
    validate_frame,
)

__all__ = [
    "MessageType",
    "RequestFrame",
    "ResponseFrame",
    "EventFrame",
    "ErrorFrame",
    "ErrorCodes",
    "create_request_frame",
    "create_response_frame",
    "create_event_frame",
    "create_error_frame",
    "validate_frame",
]



"""
Key generation and management.

Uses Ed25519 for agent identity keys.
"""

from __future__ import annotations

import base64
from dataclasses import dataclass

from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
)


@dataclass
class KeyPair:
    """Ed25519 key pair."""

    private_key: Ed25519PrivateKey
    public_key: Ed25519PublicKey
    public_key_b64: str

    def sign(self, data: bytes) -> bytes:
        """Sign data with the private key."""
        return self.private_key.sign(data)

    def verify(self, signature: bytes, data: bytes) -> bool:
        """Verify a signature. Returns False if invalid."""
        try:
            self.public_key.verify(signature, data)
            return True
        except Exception:
            return False

    def private_key_pem(self) -> str:
        """Export private key as PEM."""
        return self.private_key.private_bytes(
            Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()
        ).decode()

    def public_key_pem(self) -> str:
        """Export public key as PEM."""
        return self.public_key.public_bytes(
            Encoding.PEM, PublicFormat.SubjectPublicKeyInfo
        ).decode()


def generate_keypair() -> KeyPair:
    """Generate a new Ed25519 key pair."""
    private = Ed25519PrivateKey.generate()
    public = private.public_key()
    raw = public.public_bytes(Encoding.Raw, PublicFormat.Raw)
    b64 = base64.b64encode(raw).decode()
    return KeyPair(private_key=private, public_key=public, public_key_b64=b64)



"""
AGENIUM CLI

Simple command-line interface for agent operations.
"""

from __future__ import annotations

import argparse
import asyncio
import json
import sys

from . import __version__
from .core.types import parse_agent_uri, validate_agent_name
from .dns.resolver import DNSResolver


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="agenium",
        description="AGENIUM — Agent-to-Agent Communication SDK",
    )
    parser.add_argument("--version", action="version", version=f"agenium {__version__}")

    sub = parser.add_subparsers(dest="command")

    # resolve
    resolve_p = sub.add_parser("resolve", help="Resolve an agent:// URI")
    resolve_p.add_argument("name", help="Agent name or URI (e.g. my-agent or agent://my-agent)")
    resolve_p.add_argument("--server", default="185.204.169.26", help="DNS server")
    resolve_p.add_argument("--port", type=int, default=3000, help="DNS port")

    # validate
    validate_p = sub.add_parser("validate", help="Validate an agent name or URI")
    validate_p.add_argument("name", help="Agent name or URI to validate")

    # status
    sub.add_parser("status", help="Show SDK status")

    args = parser.parse_args()

    if args.command == "resolve":
        asyncio.run(_resolve(args))
    elif args.command == "validate":
        _validate(args)
    elif args.command == "status":
        _status()
    else:
        parser.print_help()


async def _resolve(args: argparse.Namespace) -> None:
    resolver = DNSResolver(server=args.server, port=args.port)
    try:
        agent = await resolver.resolve(args.name)
        print(json.dumps({
            "name": agent.name,
            "uri": agent.uri,
            "endpoint": agent.endpoint,
            "public_key": agent.public_key,
            "tools": [{"name": t.name, "description": t.description} for t in agent.tools],
            "capabilities": agent.capabilities,
            "ttl": agent.ttl,
        }, indent=2))
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)


def _validate(args: argparse.Namespace) -> None:
    name = args.name
    if name.startswith("agent://"):
        result = parse_agent_uri(name)
        if result:
            print(f"✓ Valid URI: agent://{result}")
        else:
            print(f"✗ Invalid URI: {name}")
            sys.exit(1)
    else:
        if validate_agent_name(name):
            print(f"✓ Valid name: {name} → agent://{name}")
        else:
            print(f"✗ Invalid name: {name}")
            sys.exit(1)


def _status() -> None:
    print(f"AGENIUM SDK v{__version__} (Python)")
    print(f"DNS Server: 185.204.169.26:3000")
    print(f"Protocol: agent://")
    print(f"Transport: HTTP/2 + mTLS")
    print(f"Docs: https://docs.agenium.net")


if __name__ == "__main__":
    main()



"""Tests for protocol types."""

from agenium.protocol.types import (
    ErrorCodes,
    ErrorFrame,
    EventFrame,
    MessageType,
    RequestFrame,
    ResponseFrame,
    create_error_frame,
    create_event_frame,
    create_request_frame,
    create_response_frame,
    validate_frame,
)


class TestFrameCreation:
    def test_request(self):
        frame = create_request_frame("tool.invoke", {"tool": "greet"})
        assert frame.type == MessageType.REQUEST
        assert frame.method == "tool.invoke"
        assert frame.params == {"tool": "greet"}
        assert frame.id

    def test_response(self):
        frame = create_response_frame("req-1", result={"ok": True})
        assert frame.type == MessageType.RESPONSE
        assert frame.request_id == "req-1"
        assert frame.result == {"ok": True}

    def test_event(self):
        frame = create_event_frame("ping", data={"ts": 123})
        assert frame.type == MessageType.EVENT
        assert frame.event == "ping"
        assert frame.data == {"ts": 123}

    def test_error(self):
        frame = create_error_frame("req-1", ErrorCodes.METHOD_NOT_FOUND, "not found")
        assert frame.type == MessageType.ERROR
        assert frame.code == -32601
        assert frame.message == "not found"


class TestValidateFrame:
    def test_valid_request(self):
        frame = create_request_frame("test")
        assert validate_frame(frame) is True

    def test_invalid_request_no_method(self):
        frame = RequestFrame()
        assert validate_frame(frame) is False

    def test_valid_response(self):
        frame = create_response_frame("req-1")
        assert validate_frame(frame) is True

    def test_invalid_response_no_request_id(self):
        frame = ResponseFrame()
        assert validate_frame(frame) is False

    def test_valid_event(self):
        frame = create_event_frame("test")
        assert validate_frame(frame) is True

    def test_invalid_not_a_frame(self):
        assert validate_frame("not a frame") is False  # type: ignore
